open Bdd;;
open Cgraph;;
open Bimap;;
open Matrix;;

(* i = I, a = ->, b = =>, A = I+->+=> , B = I+-> , C = => *)

let index2bdd bdd bi = function
  | None           -> bdd#top
  | Some (true,v)  -> bdd#var  (bi#lkup v)
  | Some (false,v) -> bdd#nvar (bi#lkup v);;

let build_r col (b:bdd_t) (g:bc_graph) (bi:int bimap_t) =
  let s = g#size                        
  and a = g#arrows                          in
  let m = matrix_empty b s                  in
  let f (i,j,c,idx) = 
      let r = if   c=col 
              then [(j,index2bdd b bi idx)]
              else []
      in  m.(i) <- sparsesum_r b r m.(i)
  in  ( List.iter f a; m );;

let build_c col (b:bdd_t) (g:bc_graph) (bi:int bimap_t) =
  let s = g#size                        
  and a = g#arrows                          in
  let m = matrix_empty b s                  in
  let f (i,j,c,idx) = 
      let r = if   c=col 
              then [(i,index2bdd b bi idx)]
              else []
      in  m.(j) <- sparsesum_r b r m.(j)
  in  ( List.iter f a; m );;

let matrices_AC g = 
  let s   = g#size                         in
  let bi  = new Bimap.bimap g#choices
  and bdd = new emptybdd (4*s*s)           in
  let ar  = build_r Green bdd g bi
  and br  = build_r Red   bdd g bi
  and bc  = build_c Red   bdd g bi
  and i   = matrix_id bdd s                in
  let mB  = sparsesum_m bdd i  ar          in
  let mA  = sparsesum_m bdd br mB            
  and mC  = bc
  in  (s,bdd,mA,mC,bi#enum);;

let matrices_BC g = 
  let s   = g#size                         in
  let bi  = new Bimap.bimap g#choices
  and bdd = new emptybdd (4*s*s)           in
  let ar  = build_r Green bdd g bi
  and br  = build_r Red   bdd g bi 
  and i   = matrix_id bdd s                in
  let mB  = sparsesum_m bdd i ar 
  and mC  = br
  in  (s,bdd,mB,mC,bi#enum);;
