open Formula;;

type pola    = Plus | Minus;;

type color   = Green | Red;;

type arrow   = int*int*color*((bool*int) option);;

type atom    = Bot | Box | Var of string ;;

type node    = Intrn of conn*pola*int*int
             | Leaf  of atom*pola
             | Extrn of atom;;

let printf = Printf.printf;;

let atom_print = function
    Bot   -> printf "."
  | Box   -> printf "%%"
  | Var s -> printf "%s" s;;

let conn_print = function 
    Conj  -> (printf "/\\")
  | Disj  -> (printf "\\/")
  | Imp   -> (printf "=>");;

let pola_print = function
    Plus  -> printf "+"
  | Minus -> printf "-";;

let arr_print (s,d,c,o) =
  begin 
  ( match c with
      Green -> printf "%d-" s
    | Red   -> printf "%d=" s ) ;
  ( match o with
      None           -> ()
    | Some (true,k)  -> printf "(%d+)" k
    | Some (false,k) -> printf "(%d-)" k ) ;
    printf ">%d" d
  end;;

let node_print = function
  | Intrn (c,p,a,b) -> (printf "("; conn_print c; printf ","; pola_print p; printf ","; printf "%d,%d)" a b)
  | Leaf  (a,p)     -> (printf "{"; atom_print a; printf ","; pola_print p; printf "}")
  | Extrn a         -> (printf "["; atom_print a; printf "]");;

(*  delta+(V)   = Xv => V,  V => top
    delta-(V)   = bot => V, V => Xv
    delta+(F)   = Xf => bot            // False
    delta-(F)   = bot => Xf
    delta+(T)   = Xt => top            // True
    delta-(T)   = top => Xt 
    delta+(A^B) = ...
*)

let arrows v top bot = 
 function
     (f,l,r,Atm a,Plus)           -> [ (f,   v a, Green, None)           ;
                                       (v a, top, Green, None)           ]
   | (f,l,r,Atm a,Minus)          -> [ (v a, f,   Green, None)           ;
                                       (bot, v a, Green, None)           ]

   | (f,l,r,False,Plus)           -> [ (f,   bot, Green, None)           ]
   | (f,l,r,False,Minus)          -> [ (bot, f,   Green, None)           ]
   | (f,l,r,True,Plus)            -> [ (f,   top, Green, None)           ]
   | (f,l,r,True,Minus)           -> [ (top, f,   Green, None)           ]

   | (f,l,r,Comp(Conj,_,_),Plus)  -> [ (f,   l,   Green, None)           ;
                                       (f,   r,   Green, None)           ]
   | (f,l,r,Comp(Conj,_,_),Minus) -> [ (l,   f,   Green, Some(true,f))   ;
                                       (r,   f,   Green, Some(false,f))  ]

   | (f,l,r,Comp(Disj,_,_),Plus)  -> [ (f,   l,   Green, Some(true,f))   ;
                                       (f,   r,   Green, Some(false,f))  ]
   | (f,l,r,Comp(Disj,_,_),Minus) -> [ (l,   f,   Green, None)           ;
                                       (r,   f,   Green, None)           ]

   | (f,l,r,Comp(Imp,_,_),Plus)   -> [ (l,   r,   Green, Some(false,f))  ;
                                       (f,   r,   Green, Some(true,f))   ]
   | (f,l,r,Comp(Imp,_,_),Minus)  -> [ (r,   f,   Green, Some(true,f))   ;
                                       (r,   l,   Red,   Some(true,f))   ;
                                       (r,   top, Red,   Some(true,f))   ;
                                       (top, f,   Green, Some(false,f))]
;;

let newpola = function 
  | (Imp,  Plus)  -> (Minus, Plus)
  | (Imp,  Minus) -> (Plus,  Minus)
  | (Conj, x)     -> (x,     x)
  | (Disj, x)     -> (x,     x);;

let convatm = function
  | Atm a -> Var a
  | False -> Bot
  | True  -> Box
  | _     -> Var "unkown error 1";;
               
let choices = Iset.collect (fun (_,_,_,x) -> match x with | None -> None | Some (_,k) -> Some k);;

class type bc_graph =
  object
    method bot      : int
    method top      : int
    method size     : int
    method vars     : int list
    method name     : int -> string
    method arrows   : arrow list
    method choices  : int list
    method node     : int -> node
  end;;

let ijlist i j = 
  let rec loop l k = 
      if   k < i 
      then l 
      else loop (k::l) (k-1)
  in loop [] j;;

class form_graph f = 
  let s   = form_size f in
  let b   = new Bimap.bimap (form_vars f) in
  let top = s+b#card+1
  and bot = s+b#card
  and sze = s+b#card+2
  and v x = s+(b#lkup x) in
  let nodes = Array.create sze (Extrn Bot) in
  let arr k l = List.append (arrows v top bot k) l in
  let rec loop l i = function
    | (p,(Comp(c,ls,rs) as f)) -> 
        let (lp,rp) = newpola (c,p)         in
        let (ll,j)  = loop l  (i+1) (lp,ls) in
        let (rl,k)  = loop ll j     (rp,rs) in
        let _       = nodes.(i) <- Intrn (c,p,i+1,j) in
        let  al     = arr (i,i+1,j,f,p) rl
        in  (al,k)
    | (p,f) -> 
        let g  = convatm f                  in
        let _  = nodes.(i) <- Leaf (g,p)    in
        let al = arr (i,-1,-1,f,p) l                                
        in  (al,i+1)                        in
  let _ = begin
            List.iter (fun x -> nodes.(v x) <- Extrn (Var x)) b#list;
            nodes.(bot) <- Extrn Bot;
            nodes.(top) <- Extrn Box
          end in
  let arrows =
    let l     = [(0,top,Red,None)] in
    let (l,_) = loop l 0 (Minus,f) 
    in  l
  in object (o)
    method bot    = bot
    method top    = top
    method size   = sze
    method vars   = ijlist s (bot-1)
    method name i = if (i>=s & i<bot) then b#enum (i-s)
                    else if i=bot     then "ff" 
                    else if i=top     then "tt"
                    else raise Not_found
    method arrows = arrows
    method choices= choices arrows
    method node   = Array.get nodes
  end;;

let print g =  
  let f i (s,_,_,_) = (i=s) 
  and h x = (arr_print x; printf ", ") 
  in  begin
        printf "This graph has %d nodes\n" g#size;
        printf "---------------------------------\n";
        for i = 0 to g#size-1 do
          printf "%2d : " i;
          node_print (g#node i);
          printf " : ";
          List.iter h (List.filter (f i) g#arrows);
          printf "\n"
        done;
        printf "---------------------------------\n"
      end;;

let dstred g = Iset.collect (fun (_,j,c,_) -> match c with | Red -> Some j | Green -> None) g#arrows;;

let condmap f l =
  let rec loop = function
        [] -> []
    | x::l -> match f x with None -> loop l | Some y -> y::(loop l)
  in loop l;;

let filterempty l = 
  let f = function [] -> false | _ -> true
  in List.filter f l;;

let cm2cond l x =
  let rec loop = function 
          []   -> false
    | (y,b)::l -> if x=y then b else loop l
  in  loop l;;

let getarrows l col =
  let cond = cm2cond l in
  let f = function
      (i,j,c,None)      -> if c=col then Some (i,j) else None
    | (i,j,c,Some(b,k)) -> if c=col & b=cond k then Some (i,j) else None
  in condmap f;;

let extract g v col = getarrows v col g#arrows;;

(* a x fixe, on veut les fleches (i,j) tries par j croissant 
   avec limination des doublons *)

(*

let getarrows g x col =
  let filt1 (i

*)

(*
  let l = match l with Some l -> l in
  let gre = getarrows l Green g#arrows
  and red = getarrows l Red   g#arrows in 
  let o = Sort.sort s gre red 
  in  filterempty (List.map (condmap g#idx2name) o);; 
*)