open Bdd;;
open Cgraph;;
open Matrix;;
open Build;;

let loglevel = ref 1;;
let log x f = if List.mem !loglevel x then f () else ();;

let print_c s c =
  begin
    Printf.printf "%4d" c.(0);
    for i = 1 to s-1 do
      Printf.printf " %4d" c.(i)
    done
  end;;

let swap a b = let k = !a in (a:=!b;b:=k);;

let log_starline ()   = Printf.printf "**********************************************************\n";;
let log_stats2 k ()   = Printf.printf " [%d]" k;;
let log_stats1 k ()   = Printf.printf "%d," k;;
let log_newline ()    = Printf.printf "\n";;
let log_column s c () = print_c s c;;
let log_flush ()      = flush_all ();;

let star_mc bdd s m b1 b2 =
  let b1r = ref b1
  and b2r = ref b2
  and k   = ref 1
  in  begin
        log [2] log_starline;
        log [2] (log_column s !b1r);
        log [2] log_newline;
        while !k>0 do
          k := prodinc_sm_c bdd m !b1r !b2r;
          swap b1r b2r;
          log [2] (log_column s !b1r);
          log [1] (log_stats1 !k);
          log [1] log_flush;
          log [2] (log_stats2 !k);
          log [2] log_newline;
        done;
        log [1] log_newline;
        log [2] log_starline;
        Array.blit !b1r 0 b1 0 s
      end;;

let ncmodel e = List.map (fun (x,b)->(e x,b));;

let fmap f l =
  let rec loop l = function
    | []    -> l
    | (x,v)::xs -> loop (try ((f x,v)::l) with
                     | Not_found -> l) xs
  in List.rev (loop [] l);;

let log_col d i () = Printf.printf "Computing column %d (trace=%d)\n" i d;;
let log_empty i () = Printf.printf "Skipping empty column %d\n" i;;
let log_res d ()   = Printf.printf "Result trace=%d\n" d;;

let trace g = 
  let (s,bdd,la,cb,e) = matrices_AC g in
  let b1 = Array.make s bdd#bot
  and b2 = Array.make s bdd#bot
  and d  = ref bdd#bot
  and bs = ref bdd
  and b  = ref bdd
  in begin
       for i = 0 to s-1 do
         match cb.(i) with
           | [] -> log [1;2] (log_empty i)
           | ci -> begin 
                     bs := !b;
                     b  := bdd#copy;
                     d  := (!b)#import !bs !d;
                     log [1;2] (log_col !d i);
                     sparseconv_sc_c !b ci b1;
                     addconst_c      !b b1 !d;
                     star_mc         !b s la b1 b2;
                     d  := (!b)#disj !d b1.(i)
                   end
       done;
       log [1;2] (log_res !d);
       match (!b)#cmodel !d with
         | None   -> None
         | Some l -> let cm  = ncmodel e l in
                     let ig  = extract g cm Green
                     and ir  = extract g cm Red in
                     let _,l = Sortg.height s ig ir
                     in  Some (fmap g#name l)
     end;;

let log_res d ()   = Printf.printf "Result sum=%d\n" d;;
  
let sum g = 
  let (s,bdd,la,lb,e) = matrices_BC g in
  let c1 = Array.make s bdd#top
  and c2 = Array.make s bdd#top
  and c3 = Array.make s bdd#top
  and d  = ref bdd#top
  and bs = ref bdd
  and b  = ref bdd
  and i  = ref 0
  and cd = ref true
  in begin
       while (bdd#istop !d) & !cd & (!i<s) do
         i  := !i+1;

         bs := !b;
         b  := bdd#copy;
         Array.iteri (fun j x -> c2.(j)<-((!b)#import !bs x)) c1;

         prod_sm_c !b   lb c2 c3;
         star_mc   !b s la c3 c1;

         cd := false;
         for j=0 to s-1 do
           if   (!b)#neq c2.(j) c3.(j)
           then cd := true
           else ()
         done;
 
         d := Array.fold_left (!b)#disj (!b)#bot c1;
         log [1;2] (log_res !d);
       done;
       Printf.printf "Sum is <top after %d iterations\n" !i;
       match (!b)#cmodel !d with
         | None   -> None
         | Some l -> let cm  = ncmodel e l in
                     let ig  = extract g cm Green
                     and ir  = extract g cm Red in
                     let _,l = Sortg.height s ig ir
                     in  Some (fmap g#name l)
     end;;