
class type equiv_t =
  object 
    method size : int
    method num  : int
    method cls  : int->int list
    method idx  : int->int
  end;;

class equiv m l =
  let a = Array.of_list l in
  let n = Array.length a in
  let b = Array.create m 0 in
  let f i j = Array.set b j i in
  let g i   = List.iter (f i) in
  let _ = Array.iteri g a 
  in object 
    method size = n
    method num  = m
    method cls i = Array.get a i
    method idx j = Array.get b j
  end;;

class type graph_t =
  object 
    method size   : int
    method add    : (int*int)->unit
    method adj    : int->int list
    method print  : unit 
    method arrows : (int*int) list
  end;;

class graph n =
  let g = Array.create n []
  in object
    method size      = n
    method add (i,j) = g.(i) <- j::g.(i)
    method adj i     = g.(i)
    method print     = 
      let f i j = (Printf.printf "%d->%d, " i j) 
      in begin
        for k = 0 to (n-1) do
          List.iter (f k) g.(k);
          Printf.printf "\n"
        done
      end
    method arrows  = 
      let l = ref [] 
      and f i j = (i,j) 
      in begin
           for k = 0 to (n-1) do
             l := List.append (List.map (f k) g.(k)) !l
           done;
           !l
         end
  end;;

let arr_add g l = List.iter g#add l;;

let arr_contract (e:equiv_t) l = 
  let f (i,j) = i<>j
  and g (i,j) = (e#idx i,e#idx j)
  in  List.filter f (List.map g l);;

let contract (g:graph_t) e =
  let k = new graph e#size in
  let l = arr_contract e (g#arrows) in  
  let _ = arr_add k l
  in  k;;

let dual (g:graph_t) =
  let k = new graph g#size in
  let f (i,j) = (j,i) 
  in  begin
        arr_add k (List.map f g#arrows);
        k
      end;;

let nfirst m = 
  let rec loop k l = 
      if   k<0 
      then l 
      else loop (k-1) (k::l) 
  in  loop (m-1) [];;

class falseenv n =
  object
    val table = Array.create n false
    method visited x = Array.get table x
    method visit   x = Array.set table x true
  end;;

let dfs ev (g:graph_t) = Dfs.depthfirst ev g#adj;;

let conn_comp g =
  let h   = dual g in
  let evg = new falseenv g#size
  and evh = new falseenv h#size in
  let a     =  dfs evg g (nfirst g#size)
  and f x l = (dfs evh h x)@l
  in  List.fold_right f a [];;

let top_sort g = 
  let ev = new falseenv g#size in
  let a  = dfs ev g (nfirst g#size)
  in  List.concat (List.rev a);;

let tsort s arrows = 
  let g = new graph s      in
  let _ = arr_add g arrows
  in  top_sort g;;

let sort s green red =
  let g   = new graph s             in
  let _   = arr_add g green         in
  let c   = conn_comp g             in
  let e   = new equiv s c           in
  let gp  = contract g e            in (* gp is the contracted of g along *) 
  let f (i,j) = (e#idx i,e#idx j)   in (* the green strongly conn. comp.  *)
  let cred    = List.map f red      in 
  let _   = arr_add gp cred         in (* adds the contracted red arrows to gp *)
  let t   = top_sort gp
  in  List.map e#cls t;;

let dualgraph s arrows =
  let f (i,j) = (j,i) 
  and g = new graph s
  in  begin
        arr_add g (List.map f arrows);
        g
      end;;

let height_upg h y x =
  let v = h.(x) 
  in  if   h.(y) < v 
      then h.(y) <- v 
      else ();;

let height_upr h y x =
  let v = h.(x)+1
  in  if   h.(y) < v 
      then h.(y) <- v 
      else ();;

let height s green red =
  let g = dualgraph s green
  and r = dualgraph s red
  and o = List.concat (sort s green red)
  and h = Array.create s 0                            in
  let m y = ( List.iter (height_upg h y) (g#adj y);
              List.iter (height_upr h y) (r#adj y) )  in
  let _ = List.iter m o                               in
  let h = Array.mapi (fun i v -> (i,v)) h             in
  let l = Array.to_list h
  in  (o,l);;

