(**
  Vector module for automaticly resizing arrays.  Uses ocamls array
  module for underlying implementation and resizes with ocamls
  append (which, IMO unintuitively concatenates two arrays).

  This module has much the same interface as ocaml's array module, but
  it uses concat to concatentate two arrays and append to append an
  element to an array.  Also, an initialization element is required
  on creation of the array, this element, is not part of the array, 
  but just a filler object for unused regions.

  @author scott
*)


(* memory allocation strategies *)
type strategy = int -> int
val const_strategy : int -> int -> int
val linear_strategy : int -> int -> int
val loglin_strategy : int -> int -> int
val default_init_sz : int
val default_strat : strategy

exception OutOfBounds of int * int

type 'a t

val create : ?sz:int -> 'a -> 'a t

val length : 'a t -> int

val get : 'a t -> int -> 'a
val unsafe_get : 'a t -> int -> 'a
val set : 'a t -> int -> 'a -> unit
val unsafe_set : 'a t -> int -> 'a -> unit


val append : 'a t -> 'a -> unit
val concat : 'a t -> 'a t -> 'a t

val sub : 'a t -> int -> int -> 'a t
val copy : 'a t -> 'a t

val fill : 'a t -> int -> int -> 'a -> unit
val blit : 'a t -> int -> 'a t -> int -> int -> unit

val to_list : 'a t -> 'a list
val of_list : 'a list -> 'a t

val iter : ('a -> unit) -> 'a t -> unit
val iteri : (int -> 'a -> unit) -> 'a t -> unit

val map : ('a -> 'b) -> 'a t -> 'b t
val mapi : (int -> 'a -> 'b) -> 'a t -> 'b t

val sort : ('a -> 'a -> int) -> 'a t -> unit
val stable_sort : ('a -> 'a -> int) -> 'a t -> unit
val fast_sort : ('a -> 'a -> int) -> 'a t -> unit
